#include "ImageCanvas.h"
#include "PropViewFrame.h"
#include <wx/dcbuffer.h>
#include <wx/settings.h>
#ifdef USE_RAW_BITMAP_SUPPORT
#	include <wx/rawbmp.h>
#else
#	include <wx/image.h>
#endif // USE_RAW_BITMAP_SUPPORT

#ifdef USE_RAW_BITMAP_SUPPORT
	typedef wxPixelData<wxBitmap, wxNativePixelFormat> PixelData;
#endif // USE_RAW_BITMAP_SUPPORT

using namespace std;

//-----------------------------------------------------------------------------
struct ImageCanvasImpl
//-----------------------------------------------------------------------------
{
	const wxBrush backgroundBrush;
	wxBitmap currentImage;
	TImageBufferPixelFormat format;
	ImageCanvasImpl() : backgroundBrush(
#ifdef __GNUC__
		wxColour(wxT("GREY")), // needed on Linux to get the same ugly grey pattern as on Windows
#else
		wxSystemSettings::GetColour( wxSYS_COLOUR_APPWORKSPACE ),
#endif
		wxCROSSDIAG_HATCH), format(ibpfRaw) {}
};

//-----------------------------------------------------------------------------
template<typename _Ty>
inline void YUV2RGB( int& r, int& g, int& b, _Ty y, _Ty u, _Ty v )
//-----------------------------------------------------------------------------
{
	b = static_cast<int>(1.164*(static_cast<double>(y)-16) + 1.596*(static_cast<double>(v)-128));
	g = static_cast<int>(1.164*(static_cast<double>(y)-16) - 0.813*(static_cast<double>(v)-128) - 0.391*(static_cast<double>(u)-128));
	r = static_cast<int>(1.164*(static_cast<double>(y)-16)                                      + 2.018*(static_cast<double>(u)-128));
}

#ifdef USE_RAW_BITMAP_SUPPORT
	//-----------------------------------------------------------------------------
	void Copy2ByteMonoImage( const ImageBuffer* pIB, PixelData::Iterator& p, PixelData& data, unsigned int shift )
	//-----------------------------------------------------------------------------
	{
		unsigned short* pSrc = reinterpret_cast<unsigned short*>(pIB->vpData);
		for( int y=0; y<pIB->iHeight; ++y )
		{
			PixelData::Iterator rowStart = p;
			char val;
			for( int x=0; x<pIB->iWidth; ++x, ++p )
			{
				val = (*pSrc) >> shift;
				++pSrc;
				p.Red() = val;
				p.Green() = val;
				p.Blue() = val;
			}
			p = rowStart;
			p.OffsetY(data, 1);
		}
	}

	//-----------------------------------------------------------------------------
	void Copy2ByteRGBImage( const ImageBuffer* pIB, PixelData::Iterator& p, PixelData& data, unsigned int shift, unsigned int inc )
	//-----------------------------------------------------------------------------
	{
		unsigned short* pSrc = reinterpret_cast<unsigned short*>(pIB->vpData);
		for( int y=0; y<pIB->iHeight; ++y )
		{
			PixelData::Iterator rowStart = p;
			char val;
			for( int x=0; x<pIB->iWidth; ++x, ++p )
			{
				p.Red() = pSrc[2] >> shift;
				p.Green() = pSrc[1] >> shift;
				p.Blue() = pSrc[0] >> shift;
				pSrc += inc;
			}
			p = rowStart;
			p.OffsetY(data, 1);
		}
	}

	//-----------------------------------------------------------------------------
	void ConvertPackedYUVToRGB( const ImageBuffer* pIB, PixelData::Iterator& p, PixelData& data )
	//-----------------------------------------------------------------------------
	{
		unsigned char* pSrc = reinterpret_cast<unsigned char*>(pIB->vpData);
		for( int line=0; line<pIB->iHeight; ++line )
		{
			int r, g, b, src_bypp = 2;
			PixelData::Iterator rowStart = p;
			unsigned char* y = ( pIB->pixelFormat == ibpfYUV422Packed ) ? pSrc                : pSrc + 1;
			unsigned char* v = ( pIB->pixelFormat == ibpfYUV422Packed ) ? pSrc + 1            : pSrc;
			unsigned char* u = ( pIB->pixelFormat == ibpfYUV422Packed ) ? pSrc + src_bypp + 1 : pSrc + src_bypp;

			for( int x=0; x<pIB->iWidth; x++, ++p )
			{
				YUV2RGB( r, g, b, *y, *u, *v );
				p.Red() = max( 0, min( r, 255 ) );
				p.Green() = max( 0, min( g, 255 ) );
				p.Blue() = max( 0, min( b, 255 ) );
				y += src_bypp;

				if( x & 1 )
				{
					u += 2 * src_bypp;
					v += 2 * src_bypp;
				}
			}
			p = rowStart;
			p.OffsetY( data, 1 );
			pSrc += pIB->pChannels[0].iLinePitch;
		}
	}

	//-----------------------------------------------------------------------------
	void ConvertPlanarYUVToRGB( const ImageBuffer* pIB, PixelData::Iterator& p, PixelData& data )
	//-----------------------------------------------------------------------------
	{
		unsigned char* pSrc = reinterpret_cast<unsigned char*>(pIB->vpData);
		for( int line=0; line<pIB->iHeight; ++line )
		{
			int r, g, b;
			PixelData::Iterator rowStart = p;
			unsigned char* y = pSrc + pIB->pChannels[0].iLinePitch * line;
			unsigned char* u = pSrc + pIB->pChannels[1].iChannelOffset + pIB->pChannels[1].iLinePitch * line;
			unsigned char* v = pSrc + pIB->pChannels[2].iChannelOffset + pIB->pChannels[2].iLinePitch * line;

			for( int x=0; x<pIB->iWidth; x++, ++p )
			{
				YUV2RGB( r, g, b, *y, *u, *v );
				p.Red() = max( 0, min( r, 255 ) );
				p.Green() = max( 0, min( g, 255 ) );
				p.Blue() = max( 0, min( b, 255 ) );
				++y;

				if( x & 1 )
				{
					++u;
					++v;
				}
			}
			p = rowStart;
			p.OffsetY( data, 1 );
		}
	}
#else
	//-----------------------------------------------------------------------------
	void Copy2ByteMonoImage( const ImageBuffer* pIB, unsigned char* pDst, unsigned int shift )
	//-----------------------------------------------------------------------------
	{
		unsigned short* pSrc = reinterpret_cast<unsigned short*>(pIB->vpData);
		unsigned char val;
		for( int i=0; i<pIB->iWidth * pIB->iHeight; i++ )
		{
			val = (*pSrc) >> shift;
			++pSrc;
			*pDst++ = val;
			*pDst++ = val;
			*pDst++ = val;
		}
	}

	//-----------------------------------------------------------------------------
	void Copy2ByteRGBImage( const ImageBuffer* pIB, unsigned char* pDst, unsigned int shift, unsigned int inc )
	//-----------------------------------------------------------------------------
	{
		unsigned short* pSrc = reinterpret_cast<unsigned short*>(pIB->vpData);
		for( int i=0; i<pIB->iWidth * pIB->iHeight; i++ )
		{
			*pDst++ = pSrc[2] >> shift;
			*pDst++ = pSrc[1] >> shift;
			*pDst++ = pSrc[0] >> shift;
			pSrc += inc;
		}
	}

	//-----------------------------------------------------------------------------
	void ConvertPackedYUVToRGB( const ImageBuffer* pIB, unsigned char* pDst )
	//-----------------------------------------------------------------------------
	{
		unsigned char* pSrc = reinterpret_cast<unsigned char*>(pIB->vpData);
		for( int line=0; line<pIB->iHeight; ++line )
		{
			int r, g, b, src_bypp = 2;
			unsigned char* y = ( pIB->pixelFormat == ibpfYUV422Packed ) ? pSrc                : pSrc + 1;
			unsigned char* v = ( pIB->pixelFormat == ibpfYUV422Packed ) ? pSrc + 1            : pSrc;
			unsigned char* u = ( pIB->pixelFormat == ibpfYUV422Packed ) ? pSrc + src_bypp + 1 : pSrc + src_bypp;

			for( int x=0; x<pIB->iWidth; x++ )
			{
				YUV2RGB( r, g, b, *y, *u, *v );
				*pDst++ = max( 0, min( r, 255 ) );
				*pDst++ = max( 0, min( g, 255 ) );
				*pDst++ = max( 0, min( b, 255 ) );
				y += src_bypp;

				if( x & 1 )
				{
					u += 2 * src_bypp;
					v += 2 * src_bypp;
				}
			}
			pSrc += pIB->pChannels[0].iLinePitch;
		}
	}

	//-----------------------------------------------------------------------------
	void ConvertPlanarYUVToRGB( const ImageBuffer* pIB, unsigned char* pDst )
	//-----------------------------------------------------------------------------
	{
		unsigned char* pSrc = reinterpret_cast<unsigned char*>(pIB->vpData);
		for( int line=0; line<pIB->iHeight; ++line )
		{
			int r, g, b;
			unsigned char* y = pSrc + pIB->pChannels[0].iLinePitch * line;
			unsigned char* u = pSrc + pIB->pChannels[1].iChannelOffset + pIB->pChannels[1].iLinePitch * line;
			unsigned char* v = pSrc + pIB->pChannels[2].iChannelOffset + pIB->pChannels[2].iLinePitch * line;
			for( int x=0; x<pIB->iWidth; x++ )
			{
				YUV2RGB( r, g, b, *y, *u, *v );
				*pDst++ = max( 0, min( b, 255 ) );
				*pDst++ = max( 0, min( g, 255 ) );
				*pDst++ = max( 0, min( r, 255 ) );
				++y;

				if( x & 1 )
				{
					++u;
					++v;
				}
			}
		}
	}
#endif // #ifdef USE_RAW_BITMAP_SUPPORT

//-----------------------------------------------------------------------------
ImageCanvas::ImageCanvas( wxWindow* pApp, wxWindow* parent, wxWindowID id /* = -1 */, const wxPoint& pos /* = wxDefaultPosition */,
	const wxSize& size /* = wxDefaultSize */, long style /* = wxSUNKEN_BORDER */, const wxString& name /* = "ImageCanvas" */, bool boActive /* = true */ )
	: DrawingCanvas( parent, id, pos, size, style, name, boActive )
//-----------------------------------------------------------------------------
{
	Init( pApp );
	m_pImpl = new ImageCanvasImpl();
}

//-----------------------------------------------------------------------------
ImageCanvas::~ImageCanvas()
//-----------------------------------------------------------------------------
{
	DeleteAOIs();
	delete m_pImpl;
}

//-----------------------------------------------------------------------------
void ImageCanvas::OnPaint( wxPaintEvent & )
//-----------------------------------------------------------------------------
{
	wxCriticalSectionLocker locker(m_critSect);
	wxPaintDC dc( this );
	DoPrepareDC( dc );

	if( !IsActive() )
	{
		return;
	}

	wxCoord clientWidth, clientHeight;
	dc.GetSize(&clientWidth, &clientHeight);
	wxMemoryDC memDC;

	dc.SetBrush( m_pImpl->backgroundBrush );
	dc.SetPen( *wxTRANSPARENT_PEN ); // no borders round rectangles

	if( !m_pImpl->currentImage.Ok() )
	{
		dc.DrawRectangle( 0, 0, clientWidth, clientHeight ); // draw background
		m_lastStartPoint.x = -1;
		m_lastStartPoint.y = -1;
		m_lastScaleFactor = -1.0;
	}
	else
	{
		const int imageHeight = m_pImpl->currentImage.GetHeight();
		const int imageWidth = m_pImpl->currentImage.GetWidth();
		memDC.SelectObject(m_pImpl->currentImage);
		if( ( imageHeight > 0 ) && ( imageWidth > 0 ) )
		{
			double scaleFactor;
			if( m_boScaleToClientSize )
			{
				const double scaleX = static_cast<double>(clientWidth) / static_cast<double>(imageWidth);
				const double scaleY = static_cast<double>(clientHeight) / static_cast<double>(imageHeight);
				// always keep the aspect ratio
				scaleFactor = ( scaleX <= scaleY ) ? scaleX : scaleY;
			}
			else
			{
				scaleFactor = m_currentZoomFactor;
			}
			const int scaledImageWidth = static_cast<int>(imageWidth * scaleFactor);
			const int scaledImageHeight = static_cast<int>(imageHeight * scaleFactor);
			int scaleCorrectedBlitXOffsetInClient = 0, scaleCorrectedBlitYOffsetInClient = 0;
			int blitXOffsetInClient = 0, blitYOffsetInClient = 0;
			int viewXStart, viewYStart;
			GetViewStart( &viewXStart, &viewYStart );
			// draw bounding box pattern with the background brush...
			if( clientWidth > scaledImageWidth )
			{
				// ... on the left and right
				const int clientRestWidth = clientWidth - scaledImageWidth;
				scaleCorrectedBlitXOffsetInClient = static_cast<int>(static_cast<double>(clientRestWidth) / (2. * scaleFactor));
				blitXOffsetInClient = static_cast<int>(clientRestWidth / 2);
				dc.DrawRectangle( 0, 0, blitXOffsetInClient, clientHeight );
				dc.DrawRectangle( scaledImageWidth + blitXOffsetInClient, 0, clientRestWidth - blitXOffsetInClient, clientHeight );
			}
			if( clientHeight > scaledImageHeight )
			{
				// ... on top and bottom
				const int clientRestHeight = clientHeight - scaledImageHeight;
				scaleCorrectedBlitYOffsetInClient = static_cast<int>(static_cast<double>(clientRestHeight) / (2. * scaleFactor));
				blitYOffsetInClient = static_cast<int>(clientRestHeight / 2);
				dc.DrawRectangle( blitXOffsetInClient, 0, scaledImageWidth, blitYOffsetInClient );
				dc.DrawRectangle( blitXOffsetInClient, scaledImageHeight + blitYOffsetInClient, scaledImageWidth, clientRestHeight - blitYOffsetInClient );
			}
			m_lastStartPoint.x = static_cast<int>(( ( blitXOffsetInClient > 0 ) ? blitXOffsetInClient : -viewXStart ) / scaleFactor);
			m_lastStartPoint.y = static_cast<int>(( ( blitYOffsetInClient > 0 ) ? blitYOffsetInClient : -viewYStart ) / scaleFactor);
			m_lastScaleFactor = scaleFactor;

			if( m_pMonitorDisplay && m_pVisiblePartOfImage )
			{
				m_pVisiblePartOfImage->m_rect = wxRect(( static_cast<int>(scaledImageWidth) > clientWidth ) ? static_cast<int>(viewXStart / scaleFactor) : 0,
				( static_cast<int>(scaledImageHeight) > clientHeight ) ? static_cast<int>(viewYStart / scaleFactor) : 0,
				( static_cast<int>(scaledImageWidth) > clientWidth ) ? static_cast<int>(clientWidth / scaleFactor) : imageWidth,
				( static_cast<int>(scaledImageHeight) > clientHeight ) ? static_cast<int>(clientHeight / scaleFactor) : imageHeight);
				m_pMonitorDisplay->Refresh( false );
			}

			dc.SetUserScale( scaleFactor, scaleFactor );
			// produces false colors when scaled and is not faster then the blit method!
			// dc.DrawBitmap( m_pImpl->currentImage, scaleCorrectedBlitXOffsetInClient, scaleCorrectedBlitYOffsetInClient, false );
			dc.Blit( scaleCorrectedBlitXOffsetInClient, scaleCorrectedBlitYOffsetInClient, imageWidth, imageHeight, &memDC, 0, 0 );
			BlitAOIs( dc, 1.0, static_cast<int>(blitXOffsetInClient / scaleFactor), static_cast<int>(blitYOffsetInClient / scaleFactor), imageWidth, imageHeight );
			dc.SetUserScale( 1.0, 1.0 );
			BlitPerformanceMessages( dc, blitXOffsetInClient, blitYOffsetInClient, m_pImpl->format );
			BlitInfoStrings( dc, scaleFactor, static_cast<int>(blitXOffsetInClient - ( viewXStart * scaleFactor )), static_cast<int>(blitYOffsetInClient - ( viewYStart * scaleFactor )), blitXOffsetInClient, blitYOffsetInClient, imageWidth, imageHeight );
			// restore old brush
			dc.SetBrush( wxNullBrush );
		}
		else
		{
			m_lastStartPoint.x = -1;
			m_lastStartPoint.y = -1;
			m_lastScaleFactor = -1.0;
			BlitInfoStrings( dc, 1.0, 10, 10, 10, 10, imageWidth, imageHeight );
		}
	}
}

//-----------------------------------------------------------------------------
void ImageCanvas::RefreshScrollbars( bool /* boMoveToMousePos = false */ )
//-----------------------------------------------------------------------------
{
	const int HScroll_Y = wxSystemSettings::GetMetric( wxSYS_HSCROLL_Y );
	const int VScroll_X = wxSystemSettings::GetMetric( wxSYS_VSCROLL_X );
	const int HScrollRange = GetScrollRange( wxHORIZONTAL );
	const int VScrollRange = GetScrollRange( wxVERTICAL );
	int clientWidth, clientHeight;
	GetClientSize( &clientWidth, &clientHeight );
	int scaledImageWidth = m_pIB ? static_cast<int>(m_pIB->iWidth * m_currentZoomFactor) : 0;
	int scaledImageHeight = m_pIB ? static_cast<int>(m_pIB->iHeight * m_currentZoomFactor) : 0;

	int virtualWidth = 0;
	int virtualHeight = 0;
	if( HScrollRange > 0 )
	{
		if( clientWidth + HScroll_Y < scaledImageWidth )
		{
			virtualWidth = scaledImageWidth;
		}
	}
	else if( clientWidth < scaledImageWidth )
	{
		virtualWidth = scaledImageWidth;
	}

	if( VScrollRange > 0 )
	{
		if( clientHeight + VScroll_X < scaledImageHeight )
		{
			virtualHeight = scaledImageHeight;
		}
	}
	else if( clientHeight < scaledImageHeight )
	{
		virtualHeight = scaledImageHeight;
	}

	if( m_boScaleToClientSize || ( ( virtualWidth == 0 ) && ( virtualHeight == 0 ) ) )
	{
		SetScrollbars( 1, 1, 0, 0 );
	}
	else
	{
		SetVirtualSize( virtualWidth, virtualHeight );
	}
}

//-----------------------------------------------------------------------------
ImageCanvas::TSaveResult ImageCanvas::SaveCurrentImage( const wxString& filenameAndPath, const wxString& extension ) const
//-----------------------------------------------------------------------------
{
	wxCriticalSectionLocker scopeLock(m_critSect);
	if( !m_pImpl->currentImage.Ok() )
	{
		return srNoImage;
	}

	wxImage img = m_pImpl->currentImage.ConvertToImage();
	if( ( extension == wxT("bmp") ) && ( m_pImpl->format == ibpfMono8 ) )
	{
		img.SetOption( wxIMAGE_OPTION_BMP_FORMAT, wxBMP_8BPP_RED );
	}

	return StoreImage( img, filenameAndPath, extension );
}

//-----------------------------------------------------------------------------
bool ImageCanvas::SetImage( const ImageBuffer* pIB )
//-----------------------------------------------------------------------------
{
	wxCriticalSectionLocker locker( m_critSect );

	if( !IsActive() || !pIB )
	{
		return true;
	}

	m_pIB = pIB;

	bool boImageSizeChanged = false;
	if( !m_pImpl->currentImage.Ok() || ( m_pImpl->currentImage.GetWidth() != pIB->iWidth ) ||
		( m_pImpl->currentImage.GetHeight() != pIB->iHeight ) )
	{
		boImageSizeChanged = true;
		SetMaxZoomFactor( pIB, std::max( m_pImpl->currentImage.GetWidth(), m_pImpl->currentImage.GetHeight() ) );
	}

	if( !pIB->vpData )
	{
		// restore initial status as no valid image is currently available
		m_pImpl->currentImage = wxBitmap();
		Refresh( false );
		return true;
	}
#ifdef USE_RAW_BITMAP_SUPPORT
	if( boImageSizeChanged )
	{
		m_pImpl->currentImage = wxBitmap( pIB->iWidth, pIB->iHeight, 24 );
	}
	PixelData data(m_pImpl->currentImage);
	PixelData::Iterator p(data);
	switch( pIB->pixelFormat )
	{
	case ibpfMono8:
		{
			unsigned char* pSrc = static_cast<unsigned char*>(pIB->vpData);
			for( int y=0; y<pIB->iHeight; y++ )
			{
				PixelData::Iterator rowStart = p;
				for( int x=0; x<pIB->iWidth; x++, p++ )
				{
					p.Red() = *pSrc;
					p.Green() = *pSrc;
					p.Blue() = *pSrc++;
				}
				p = rowStart;
				p.OffsetY(data, 1);
			}
		}
		break;
	case ibpfMono16:
		Copy2ByteMonoImage( pIB, p, data, 8 );
		break;
	case ibpfRGBx888Packed:
		{
			unsigned char* pSrc = static_cast<unsigned char*>(pIB->vpData);
			for( int y=0; y<pIB->iHeight; y++ )
			{
				PixelData::Iterator rowStart = p;
				for( int x=0; x<pIB->iWidth; x++, p++ )
				{
					p.Blue() = *pSrc++;
					p.Green() = *pSrc++;
					p.Red() = *pSrc++;
					++pSrc;
				}
				p = rowStart;
				p.OffsetY(data, 1);
			}
		}
		break;
	case ibpfRGB888Packed:
		{
			unsigned char* pSrc = static_cast<unsigned char*>(pIB->vpData);
			for( int y=0; y<pIB->iHeight; y++ )
			{
				PixelData::Iterator rowStart = p;
				for( int x=0; x<pIB->iWidth; x++, p++ )
				{
					p.Blue() = *pSrc++;
					p.Green() = *pSrc++;
					p.Red() = *pSrc++;
				}
				p = rowStart;
				p.OffsetY(data, 1);
			}
		}
		break;
	case ibpfYUV422Packed:
	case ibpfYUV422_UYVYPacked:
		ConvertPackedYUVToRGB( pIB, p, data );
		break;
	case ibpfYUV422Planar:
		ConvertPlanarYUVToRGB( pIB, p, data );
		break;
	case ibpfRGBx888Planar:
		{
			unsigned char* pr = static_cast<unsigned char*>(pIB->vpData);
			unsigned char* pg = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[1].iChannelOffset;
			unsigned char* pb = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[2].iChannelOffset;
			for( int y=0; y<pIB->iHeight; y++ )
			{
				PixelData::Iterator rowStart = p;
				for( int x=0; x<pIB->iWidth; x++, p++ )
				{
					p.Red() = *pr++;
					p.Green() = *pg++;
					p.Blue() = *pb++;
				}
				p = rowStart;
				p.OffsetY(data, 1);
			}
		}
		break;
	case ibpfMono10:
		Copy2ByteMonoImage( pIB, p, data, 2 );
		break;
	case ibpfMono12:
		Copy2ByteMonoImage( pIB, p, data, 4 );
		break;
	case ibpfMono12Packed_V2:
		for( int y=0; y<pIB->iHeight; y++ )
		{
			unsigned char* pSrcLineStart = static_cast<unsigned char*>(pIB->vpData) + y*pIB->pChannels[0].iLinePitch;
			PixelData::Iterator rowStart = p;
			for( int x=0; x<pIB->iWidth; x+=2 )
			{
				p.Red() = *pSrcLineStart;
				p.Green() = *pSrcLineStart;
				p.Blue() = *pSrcLineStart;
				pSrcLineStart +=2;
				++p;
				p.Red() = *pSrcLineStart;
				p.Green() = *pSrcLineStart;
				p.Blue() = *pSrcLineStart++;
			}
			p = rowStart;
			p.OffsetY(data, 1);
		}
		break;
	case ibpfMono14:
		Copy2ByteMonoImage( pIB, p, data, 6 );
		break;
	case ibpfRGB101010Packed:
		Copy2ByteRGBImage( pIB, p, data, 2, 3 );
		break;
	case ibpfRGB121212Packed:
		Copy2ByteRGBImage( pIB, p, data, 4, 3 );
		break;
	case ibpfRGB141414Packed:
		Copy2ByteRGBImage( pIB, p, data, 6, 3 );
		break;
	case ibpfRGB161616Packed:
		Copy2ByteRGBImage( pIB, p, data, 8, 3 );
		break;
	default:
		return false;
		break;
	}
#else
	wxImage img( pIB->iWidth, pIB->iHeight, false );
	switch( pIB->pixelFormat )
	{
	case ibpfMono8:
		{
			unsigned char* pDst = img.GetData();
			unsigned char* pSrc = static_cast<unsigned char*>(pIB->vpData);
			for( int i=0; i<pIB->iWidth * pIB->iHeight; i++ )
			{
				*pDst++ = *pSrc;
				*pDst++ = *pSrc;
				*pDst++ = *pSrc++;
			}
		}
		break;
	case ibpfMono16:
		Copy2ByteMonoImage( pIB, img.GetData(), 8 );
		break;
	case ibpfRGBx888Packed:
		{
			unsigned char* pDst = img.GetData();
			unsigned char* pSrc = static_cast<unsigned char*>(pIB->vpData);
			for( int i=0; i<pIB->iWidth * pIB->iHeight; i++ )
			{
				*pDst++ = pSrc[2];
				*pDst++ = pSrc[1];
				*pDst++ = *pSrc;
				pSrc += 4;
			}
		}
		break;
	case ibpfRGB888Packed:
		{
			unsigned char* pDst = img.GetData();
			unsigned char* pSrc = static_cast<unsigned char*>(pIB->vpData);
			for( int i=0; i<pIB->iWidth * pIB->iHeight; i++ )
			{
				*pDst++ = pSrc[2];
				*pDst++ = pSrc[1];
				*pDst++ = *pSrc;
				pSrc += 3;
			}
		}
		break;
	case ibpfYUV422Packed:
	case ibpfYUV422_UYVYPacked:
		ConvertPackedYUVToRGB( pIB, img.GetData() );
		break;
	case ibpfYUV422Planar:
		ConvertPlanarYUVToRGB( pIB, img.GetData() );
		break;
	case ibpfRGBx888Planar:
		{
			unsigned char* pDst = img.GetData();
			unsigned char* pr = static_cast<unsigned char*>(pIB->vpData);
			unsigned char* pg = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[1].iChannelOffset;
			unsigned char* pb = static_cast<unsigned char*>(pIB->vpData) + pIB->pChannels[2].iChannelOffset;
			for( int i=0; i<pIB->iWidth * pIB->iHeight; i++ )
			{
				*pDst++ = *pr++;
				*pDst++ = *pg++;
				*pDst++ = *pb++;
			}
		}
		break;
	case ibpfMono10:
		Copy2ByteMonoImage( pIB, img.GetData(), 2 );
		break;
	case ibpfMono12:
		Copy2ByteMonoImage( pIB, img.GetData(), 4 );
		break;
	case ibpfMono12Packed_V2:
		{
			unsigned char* pDst = img.GetData();
			for( int y=0; y<pIB->iHeight; y++ )
			{
				unsigned char* pSrcLineStart = static_cast<unsigned char*>(pIB->vpData) + y*pIB->pChannels[0].iLinePitch;
				for( int x=0; x<pIB->iWidth; x+=2 )
				{
					*pDst++ = *pSrcLineStart;
					*pDst++ = *pSrcLineStart;
					*pDst++ = *pSrcLineStart;
					pSrcLineStart +=2;
					*pDst++ = *pSrcLineStart;
					*pDst++ = *pSrcLineStart;
					*pDst++ = *pSrcLineStart++;
				}
			}
		}
		break;
	case ibpfMono14:
		Copy2ByteMonoImage( pIB, img.GetData(), 6 );
		break;
	case ibpfRGB101010Packed:
		Copy2ByteRGBImage( pIB, img.GetData(), 2, 3 );
		break;
	case ibpfRGB121212Packed:
		Copy2ByteRGBImage( pIB, img.GetData(), 4, 3 );
		break;
	case ibpfRGB141414Packed:
		Copy2ByteRGBImage( pIB, img.GetData(), 6, 3 );
		break;
	case ibpfRGB161616Packed:
		Copy2ByteRGBImage( pIB, img.GetData(), 8, 3 );
		break;
	default:
		return false;
		break;
	}
	m_pImpl->currentImage = wxBitmap( img, -1 );
#endif // USE_RAW_BITMAP_SUPPORT
	m_pImpl->format = pIB->pixelFormat;
	if( boImageSizeChanged )
	{
		RefreshScrollbars();
	}
	Refresh( boImageSizeChanged );
	return true;
}

